'''
Date: 2024-08-19 10:41:56
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-15 19:27:13
Description: 
'''
import os
from functools import partial
from pathlib import Path
from threading import Thread

from pymol import cmd


class GUILauncher:
    def __init__(self, app = None, n_threads=4):
        self.host = Thread(name = 'GUI-Host', target=self.init_run_self_in_background,
                           kwargs = dict(n_threads=n_threads), daemon=False)
        self.host.start()
        
    def init_run_self_in_background(self, n_threads: int):
        print('Lazydock pymol Plugin: initializing GUI plugin...')
        from mbapy_lite.web import TaskPool
        from nicegui import app, ui

        from lazydock.pml_plugin.lazy_dlg import LazyDLG
        from lazydock.pml_plugin.lazy_plot import LazyPlot
        from lazydock.pml_plugin.lazy_pml import LazyPml
        from lazydock.pml_plugin.lazy_pocket import LazyPocket
        
        self._now_molecule = cmd.get_names_of_type('object:molecule') or []
        self._now_selection = cmd.get_names_of_type('selection') + ['sele']
        self.ui_update_func = []
        ui.timer(1, self.ui_update_content_from_pymol)
        
        self.lazy_pml = LazyPml(self)
        self.lazy_pocket = LazyPocket(self)
        self.lazy_dlg = LazyDLG(self)
        self.lazy_plot = LazyPlot(self)
        
        self.build_gui()
        
        self.taskpool = TaskPool('threads', n_threads).start()
        print('Lazydock pymol Plugin: GUI plugin initialized.')
        ui.run(title='LazyDock', host='localhost', port=8090, reload=False)
        
        
    def ui_update_content_from_pymol(self, ):
        self._now_molecule = cmd.get_names_of_type('object:molecule')
        self._now_selection = cmd.get_names_of_type('selection') + ['sele']
        
        for fn in self.ui_update_func:
            fn()
            
    def load_gpf(self, content: str):
        from lazydock.pml.thirdparty.draw_bounding_box import draw_box
        lines = content.split('\n')
        grid_size_line = list(filter(lambda x: x.startswith('npts'), lines))[0]
        grid_center_line = list(filter(lambda x: x.startswith('gridcenter'), lines))[0]
        scale_line = list(filter(lambda x: x.startswith('spacing'), lines))[0]
        scale = float(scale_line.split(' ')[1])
        getter_fn = lambda line, scale: list(map(lambda x: x*scale, map(float, line.split(' ')[1:4])))
        grid_center = getter_fn(grid_center_line, 1)
        grid_size = getter_fn(grid_size_line, scale)
        min_x, min_y, min_z = [grid_center[i] - grid_size[i] / 2 for i in range(3)]
        max_x, max_y, max_z = [grid_center[i] + grid_size[i] / 2 for i in range(3)]
        draw_box(min_x, min_y, min_z, max_x, max_y, max_z)

    async def uni_load(self, event):
        from mbapy_lite.file import decode_bits_to_str, opts_file
        from nicegui import app, ui
        
        if event is None:
            from lazydock.pml_plugin._nicegui.local_file_picker import \
                local_file_picker
            paths = await local_file_picker(os.path.abspath('.'), upper_limit=None,
                                            multiple=True, show_hidden_files=True,
                                            file_extensions=['.pdb', '.pdbqt', '.dlg'])
            if paths is None:
                return ui.notify('No file selected')
            for path in paths:
                if path.endswith('.pdb') or path.endswith('.pdbqt'):
                    cmd.load(path)
                elif path.endswith('.dlg'):
                    await self.lazy_dlg.pose_page.load_dlg_file(None, path=path)
                elif path.endswith('.gpf'):
                    self.load_gpf(opts_file(path))
        else:
            if event.name.endswith('.pdb') or event.name.endswith('.pdbqt'):
                pdbstr = decode_bits_to_str(event.content.read())
                cmd.read_pdbstr(pdbstr, Path(event.name).stem)
            elif event.name.endswith('.dlg'):
                await self.lazy_dlg.pose_page.load_dlg_file(event, path=None)
            elif event.name.endswith('.gpf'):
                self.load_gpf(decode_bits_to_str(event.content.read()))
                    
    def build_gui(self):
        from nicegui import app, ui
        with ui.header(elevated=True).style('background-color: #3874c8'):
            ui.label('LazyDock | Pymol Plugin').classes('text-h4')
            ui.space()
            ui.upload(label='upload', auto_upload=True, on_upload=self.uni_load).style('max-width: 300px;max-height: 35px')
            ui.button('Open', on_click=partial(self.uni_load, event=None), icon='file_open')
            ui.button('Exit', on_click=app.shutdown, icon='power')
        with ui.splitter(value=10).classes('w-full h-full') as splitter:
            with splitter.before:
                with ui.tabs().props('vertical align=left active-bg-color=blue').classes('w-full') as tabs:
                    lazy_pml_tab = ui.tab('Pymol')
                    lazy_pocket_tab = ui.tab('Pocket')
                    lazy_dlg_tab = ui.tab('DLG')
                    lazy_plot_tab = ui.tab('Plot')
            with splitter.after:
                with ui.tab_panels(tabs, value=lazy_pml_tab) \
                        .props('vertical').classes('w-full h-full'):
                    with ui.tab_panel(lazy_pml_tab):
                        self.lazy_pml.build_gui()
                    with ui.tab_panel(lazy_pocket_tab):
                        self.lazy_pocket.build_gui()
                    with ui.tab_panel(lazy_dlg_tab):
                        self.lazy_dlg.build_gui()
                    with ui.tab_panel(lazy_plot_tab):
                        self.lazy_plot.build_gui()


def exec_from_cli():
    app = GUILauncher()

    
if __name__ in {"__main__", "__mp_main__"}:
    # dev code
    app = GUILauncher()


import importlib
import os
import sys

from mbapy_lite.base import get_fmt_time
from mbapy_lite.file import opts_file

if __name__ == '__main__':
    from lazydock.utils import get_storage_path
else:
    from ..utils import get_storage_path

scripts_info = opts_file(get_storage_path('lazydock-cli-scripts-list.json'), way = 'json')
exec2script = {exec_name: script_info['script name'] for script_info in scripts_info.values() for exec_name in script_info['exec_names']}


def print_version_info():
    import lazydock
    print('lazydock python package command-line tools')
    print('lazydock version: ', lazydock.__version__, ', build: ', lazydock.__build__)
    print('lazydock author: ', lazydock.__author__, ', email: ', lazydock.__author_email__)
    print('lazydock url: ', lazydock.__url__, ', license: ', lazydock.__license__)
    
def print_help_info():
    help_info = """
            usage-1: lazydock-cli [-h] [-l | -i]
            options:
            -h, --help  show this help message and exit
            -l, --list  print scripts list
            -i, --info  print scripts info
            usage-2: lazydock-cli [sub-scripts-name] [args] [-h]
            options:
            sub-scripts-name  name of scripts in lazydock.scripts
            args  args for sub-scripts
            -h, --help  show this help message and exit
            """
    print(help_info)
    
def print_scripts_list():
    for idx, script in enumerate(scripts_info):
        print(f'scripts {idx:3d}: {scripts_info[script]["name"]}')
        print(f'script file name: {scripts_info[script]["script name"]}')
        print(f'exec names: {", ".join(scripts_info[script]["exec_names"])}')
        print(scripts_info[script]['brief'])
        print('-'*100)

def print_scripts_info():
    for idx, script in enumerate(scripts_info):
        print(f'scripts {idx:3d}: {scripts_info[script]["name"]}')
        print(f'script file name: {scripts_info[script]["script name"]}')
        print(f'exec names: {", ".join(scripts_info[script]["exec_names"])}')
        print(scripts_info[script]['brief'])
        print(scripts_info[script]['detailed'])
        print('-'*100)
        
def exec_scripts():
    import lazydock

    # check --pause-after-exec argumet
    pause_after_exec = '--pause-after-exec' in sys.argv
    if pause_after_exec:
        sys.argv.remove('--pause-after-exec')
    # check and exec scripts NOTE: DO NOT use exec
    script_name = exec2script[sys.argv[1]]
    script = importlib.import_module(f'.{script_name}', 'lazydock.scripts')
    script.main(sys.argv[2:])
    # pause if --pause-after-exec
    if pause_after_exec:
        os.system('pause') # avoid cmd window close immediately
    
def main():  
    def _handle_unkown():
        print(f'lazydock-cli: unkown scripts: {sys.argv[1]} and args: {", ".join(sys.argv[2:])}, SKIP.')
        print('bellow are all scripts list:\n\n')
        print_scripts_list()
        
    if len(sys.argv) == 1:
        print_version_info()
    elif len(sys.argv) == 2:
        if sys.argv[1] in ['-l', '--list']:
            print_scripts_list()
        elif sys.argv[1] in ['-i', '--info']:
            print_scripts_info()
        elif sys.argv[1] in ['-h', '--help']:
            print_help_info()
        elif sys.argv[1] in exec2script:
            # exec scripts with only ZERO arg
            exec_scripts()
        elif os.path.exists(sys.argv[1]) and sys.argv[1].endswith('.mpss'):
            # load lazydock Script Session file and exec
            from lazydock.scripts._script_utils_ import Command
            print(f'loading session from file: {sys.argv[1]}')
            Command(None).exec_from_session(sys.argv[1])
            os.system('pause') # avoid cmd window close immediately
        else:
            _handle_unkown()
    else:
        if sys.argv[1] in exec2script:
            # exec scripts
            exec_scripts()
        else:
            _handle_unkown()
    # exit
    print(f'\nlazydock-cli: exit at {get_fmt_time()}')
            

if __name__ == '__main__':
    main()


from __future__ import absolute_import, print_function

import re
import time
import tkinter as tk
import tkinter.filedialog as tkFileDialog
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from mbapy_lite.base import put_err
from mbapy_lite.file import decode_bits_to_str, opts_file
from mbapy_lite.game import BaseInfo
from mbapy_lite.stats.cluster import KMeans, KMeansBackend
from mbapy_lite.web_utils.task import TaskPool
from pymol import cmd
from tqdm import tqdm

from lazydock.utils import uuid4

# atom-type, atom-number, atom-name, residue-name, chain-name, residue-number, x, y, z, occupancy, temperature-factor
# ATOM      1  CA  LYS     7     136.747 133.408 135.880 -0.06 +0.10 OA
# HETATM   79  CB  PRO     9     132.763 138.092 170.987+13.19 -0.02    +0.037 C  \n
# HETATM   40  H   UNK     1     -25.811   4.632  -7.417  1.00 20.00     0.163 HD
PDB_PATTERN = r"(ATOM|HETATM) +(\d+) +(\w+) +(\w+) +(\w+)? +(\d+) +([\d\-\.]+) +([\d\-\.]+) +([\d\-\.]+)([ \+\-]+[\d\-\.]+)([ \+\-]+[\d\-\.]+) [ \.\-\+\d]+ ([A-Z]+)"
PDB_FORMAT = "{:6s}{:>5s}  {:<3s} {:>3s} {:1s}{:>4s}    {:>8s}{:>8s}{:>8s}{:>6s}{:>6s}          {:>2s}  "
PDB_FORMAT2= "{:6s}{:>5s} {:<4s} {:>3s} {:1s}{:>4s}    {:>8s}{:>8s}{:>8s}{:>6s}{:>6s}          {:>2s}  "


def calcu_RMSD(pose1: Union[str, 'ADModel'], pose2: Union[str, 'ADModel'], _cmd = None):
    _cmd = _cmd or cmd
    def _get_pose(pose):
        if isinstance(pose, ADModel):
            name = uuid4()
            _cmd.read_pdbstr(pose.as_pdb_string(), name)
            return name, True
        return pose, False
    pml_pose1, is_loaded1 = _get_pose(pose1)
    pml_pose2, is_loaded2 = _get_pose(pose2)
    rmsd = _cmd.rms(pml_pose1, pml_pose2, cycles=0) # cutoff = float: outlier rejection cutoff (only if cycles>0) {default: 2.0}
    if is_loaded1:
        _cmd.delete(pml_pose1)
    if is_loaded2:
        _cmd.delete(pml_pose2)
    return rmsd


class ADModel(BaseInfo):
    """STORAGE CLASS FOR DOCKED LIGAND"""
    def __init__(self, content: str = None, _sort_atom_by_res: bool = False,
                 _parse2std: bool = False, default_chain: str = 'Z'):
        self.energy = 0.
        self.name = ''
        self.run_idx = 0
        self.info = ''
        self.pdb_string = ''
        self.default_chain = default_chain
        self.pdb_lines = []
        self.coords = None
        self.coords_only_bb = None
        if content is not None:
            self.parse_content(content, _sort_atom_by_res, _parse2std)

    def parse_content(self, content: str, _sort_atom_by_res: bool = False,
                      _parse2std: bool = False):
        # parse pdb lines
        self.info = content
        self.pdb_lines = list(map(lambda x: x[0], re.findall(r'((ATOM|HETATM).+?\n)', self.info)))
        self.pdb_atoms = list(map(list, re.findall(PDB_PATTERN, self.info)))
        if len(self.pdb_atoms) != len(self.pdb_lines):
            put_err(f'pdb_atoms and pdb_lines length not match: {len(self.pdb_atoms)} vs {len(self.pdb_lines)}, just use pdb_lines')
            _sort_atom_by_res = _parse2std = False
        if _sort_atom_by_res:
            pack = sorted(zip(self.pdb_lines, self.pdb_atoms), key = lambda x : (x[1][4], int(x[1][5]), int(x[1][1])))
            self.pdb_lines, self.pdb_atoms = zip(*pack)
        if _parse2std:
            for atom in self.pdb_atoms:
                if not atom[4]:
                    atom[4] = self.default_chain
                if len(atom[2]) == 4:
                    atom[-1] = atom[2][1] # such as 1HH1
                elif len(atom[2]) >= 1:
                    if atom[2][0].isdigit():
                        atom[-1] = atom[2][1] # such as 1HH
                    else:
                        atom[-1] = atom[2][0] # such as H, CA, NH1
            self.pdb_lines = [PDB_FORMAT2.format(*line) if len(line[2])==4 else PDB_FORMAT.format(*line) for line in self.pdb_atoms]
            self.pdb_string = '\n'.join(self.pdb_lines)
        else:
            self.pdb_string = ''.join(self.pdb_lines)
        # parse energy could be str.find?
        energy_line = re.findall(r'USER.+?Free Energy of Binding.+?\n', self.info)
        if energy_line:
            entr = energy_line[0].split('=')[1]
            self.energy = float(entr.split()[0])
        else:
            energy_line = re.findall(r'REMARK.+?VINA RESULT.+?\n', self.info)
            if energy_line:
                entr = energy_line[0].split(':')[1]
                self.energy = float(entr.split()[0])
            else:
                self.energy = None
        # parse run idx
        self.run_idx = int(re.findall(r'MODEL +(\d+)', self.info)[0])

    def as_pdb_string(self):
        return self.pdb_string

    def info_string(self):
        return self.info
    
    def parse_coords(self, only_bb: bool = True):
        if self.coords is None or only_bb != self.coords_only_bb:
            self.coords_only_bb = only_bb
            self.coords = np.array([list(map(float, line[6:9])) for line in self.pdb_atoms if line[2] in ['C', 'N', 'O', 'CA'] or not only_bb])
        if self.coords.shape[0] == 0:
            # considerate that atom name may not be 'C', 'N', 'O', 'CA', try to use last column
            self.coords = np.array([list(map(float, line[6:9])) for line in self.pdb_atoms if line[-1] in ['C', 'N', 'O', 'CA'] or not only_bb])
        return self.coords
    
    def calcu_rmsd_by_coords(self, other: 'ADModel', only_bb: bool = True):
        self_coords = self.parse_coords(only_bb)
        other_coords = other.parse_coords(only_bb)
        d2 = np.sum((self_coords - other_coords)**2, axis=1)
        return np.sqrt(np.mean(d2))
    
    def rmsd(self, other: 'ADModel', backend: str = 'pymol', only_bb: bool = True):
        if backend == 'pymol':
            return calcu_RMSD(self, other)
        elif backend == 'rdkit':
            from rdkit import Chem
            from rdkit.Chem import AllChem
            pose1 = Chem.MolFromPDBBlock(self.as_pdb_string(), removeHs=True, sanitize=False)
            pose2 = Chem.MolFromPDBBlock(other.as_pdb_string(), removeHs=True, sanitize=False)
            return AllChem.GetBestRMS(pose1, pose2)
        elif backend == 'numpy':
            return self.calcu_rmsd_by_coords(other, only_bb)
        else:
            raise ValueError(f'Unsupported backend: {backend}')


class RmsdClusterBackend(KMeansBackend):
    def __init__(self) -> None:
        super().__init__('scipy')
    def cdist(self, data, centers):
        return np.sqrt(np.mean(np.sum((data[:, None] - centers[None, :])**2, axis=-1), axis=-1))


class DlgFile(BaseInfo):
    def __init__(self, path: str = None, content: str = None,
                 sort_pdb_line_by_res: bool = False,
                 parse2std: bool = False):
        # load content from file if path is provided
        self.path = path
        self.sort_pdb_line_by_res = sort_pdb_line_by_res
        self.parse2std = parse2std
        if path is not None:
            self.content = decode_bits_to_str(opts_file(path, 'rb'))
        elif content is not None:
            self.content = content
        else:
            self.content = None
        # decode content to pose_lst
        if self.content is not None:
            self.pose_lst: List[ADModel] = self.decode_content()
            self.sort_pose()
        else:
            self.pose_lst: List[ADModel] = []
        self.n2i = {}
        
    def __len__(self):
        return len(self.pose_lst)
    
    def merge(self, other: 'DlgFile', inplace: bool = True):
        """only merge poses, not content or other attributes"""
        if not inplace:
            new_dlg = DlgFile()
            new_dlg.pose_lst = self.pose_lst + other.pose_lst
            return new_dlg
        else:
            self.pose_lst += other.pose_lst
            return self
        
    def sort_pose(self, key: Callable[[ADModel], Any] = None,
                  inplace: bool = True, reverse: bool = False) -> List[ADModel]:
        if key is None:
            key = lambda x : x.energy
        if inplace:
            self.pose_lst.sort(key=key, reverse=reverse)
            ret = self.pose_lst
        else:
            ret = sorted(self.pose_lst, key=key, reverse=reverse)
        return ret
        
    def decode_content(self):
        dlg_pose_lst = []
        for model in re.findall('MODEL.+?ENDMDL', self.content, re.DOTALL):
            model = model.replace('\nDOCKED: ', '\n')
            dlg_pose_lst.append(ADModel(model, self.sort_pdb_line_by_res, self.parse2std))
        return dlg_pose_lst
    
    def asign_pose_name(self, pose_names: List[str]):
        if len(pose_names) != len(self.pose_lst):
            raise ValueError("Number of pose names must match number of poses")
        for i, name in enumerate(pose_names):
            self.n2i[name] = i
        
    def asign_prop(self, prop: str, value: List[Any]):
        if value is not None and len(value) == len(self.pose_lst):
            setattr(self, prop, value)
                
    def set_pose_prop(self, prop: str, value: Any, pose_name: str = None, pose_idx: int = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        if not hasattr(self, prop):
            setattr(self, prop, [None]*len(self.pose_lst))
        getattr(self, prop)[pose_idx] = value
            
    def get_pose(self, pose_name: str = None, pose_idx: int = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        return self.pose_lst[pose_idx]
    
    def get_pose_prop(self, prop: str, pose_name: str = None, pose_idx: int = None, default: Any = None):
        if pose_name is None and pose_idx is None:
            raise ValueError("Either pose_name or pose_idx must be provided")
        if pose_name is not None:
            pose_idx = self.n2i[pose_name]
        if hasattr(self, prop):
            return getattr(self, prop)[pose_idx]
        else:
            return default
    
    def parse_energies(self):
        """
        calculate energies of all poses, and sort them by idx and energy
        
        - self.sorted_energies_by_idx: [N_pose], sorted energies by pose run index
        - self.sorted_energies: [N_pose], sorted energies in descending order, first is the biggest
        - self.pooled_energies_by_idx: [N_pose], pooled energies by pose index
        - self.mean_energy: float, mean energy of all poses
        """
        idx_energy = [[pose.run_idx, pose.energy] for pose in self.pose_lst]
        self.sorted_energies_by_idx = list(map(lambda x: x[1], sorted(idx_energy, key=lambda x: x[0])))
        self.sorted_energies = sorted(self.sorted_energies_by_idx, reverse=True)
        self.pooled_energies_by_idx = [min(self.sorted_energies_by_idx[:i]) for i in range(1, len(self.sorted_energies_by_idx)+1)]
        self.mean_energy = np.mean(self.sorted_energies)
        
    def plot_energy_curve(self, figsize: Tuple[int, int] = (8, 6)):
        """
        plot energy curve of all pose
        - plot pooled energy curve
        - plot individual energy curve
        - plot hist of individual energy
        - plot mean energy
        - plot best energy
        
        Returns:
            - fig, ax, ax_histy
        """
        fig = plt.figure(figsize=figsize)
        gs = fig.add_gridspec(1, 2,  width_ratios=(4, 1), left=0.1, right=0.9, bottom=0.1, top=0.9,
                wspace=0.05, hspace=0.05)
        # Create the Axes.
        ax = fig.add_subplot(gs[0, 0])
        ax_histy = fig.add_subplot(gs[0, 1], sharey=ax)
        # Draw the scatter plot and marginals.
        ax.plot(self.pooled_energies_by_idx, linewidth=2, label='Minimum Energy in Run Order')
        ax.plot([self.mean_energy] * len(self.pooled_energies_by_idx), c='black', linewidth=2, label=f'Mean Energy: {self.mean_energy:.4f} kcal/mol')
        ax.plot([self.sorted_energies[-1]] * len(self.pooled_energies_by_idx), c='gray', linestyle='--',
                linewidth=2, label=f'Best Energy: {self.sorted_energies[-1]:.4f} kcal/mol')
        ax.scatter(list(range(len(self.sorted_energies_by_idx))), self.sorted_energies_by_idx,
                    alpha=0.4, c='green', s=50, label='Individual Energy in Run Order')
        ax.scatter(list(range(len(self.sorted_energies))), self.sorted_energies,
                    alpha=0.2, c='red', s=30, label='Individual Energy in Descending Order')
        ax.set_xlabel('Pose Index', fontdict={'size': 14})
        ax.set_ylabel('Energy (kcal/mol)', fontdict={'size': 14})
        ax.set_title(f'Energy Curve for {Path(self.path).stem}', fontdict={'size': 16})
        # plot hist
        ax_histy.tick_params(axis="y", labelleft=False)
        ax_histy.hist(self.sorted_energies, bins=int(max(self.sorted_energies)-min(self.sorted_energies)), orientation='horizontal')
        # minor works
        ax.tick_params(axis='both', which='major', labelsize=12)
        ax_histy.tick_params(axis='both', which='minor', labelsize=12)
        ax.legend(fontsize=12, draggable=True, framealpha=0.8)
        plt.tight_layout()
        return fig, ax, ax_histy
        
    def rmsd(self, backend: str = 'pymol', taskpool: TaskPool = None, verbose: bool = False):
        # if backend is numpy, use matrix calculation to calculate rmsd for acceleration
        if backend == 'numpy':
            # [N_pose, M_atom, D_3]
            coords = np.concatenate([[pose.parse_coords()] for pose in self.pose_lst], axis=0)
            # d2 = [N_pose, N_pose, M_atom, D_3] => [N_pose, N_pose, M_atom]: np.sum((coords[None, :, :] - coords[:, None, :])**2, axis=-1)
            # sum+devide => mean => [N_pose, N_pose, M_atom] => [N_pose, N_pose]
            return np.sqrt(np.mean(np.sum((coords[None, :, :] - coords[:, None, :])**2, axis=-1), axis=-1))
        # if is other backend, loop over all pairs and calculate rmsd
        rmsd_mat = np.zeros((len(self.pose_lst), len(self.pose_lst))).tolist()
        for i, pose_i in tqdm(enumerate(self.pose_lst), desc='Calculating RMSD', total=len(self.pose_lst), disable=not verbose):
            for j, pose_j in enumerate(self.pose_lst):
                if i > j:
                    continue
                if taskpool is None:
                    rmsd_mat[i][j] = pose_i.rmsd(pose_j, backend=backend)
                else:
                    while taskpool.count_waiting_tasks() > 0:
                        time.sleep(0.1)
                    rmsd_mat[i][j] = taskpool.add_task(f'{i}-{j}', pose_i.rmsd, pose_j, backend=backend)
        for i in range(len(rmsd_mat)):
            for j in range(len(rmsd_mat)):
                if i > j:
                    rmsd_mat[i][j] = rmsd_mat[j][i]
                elif taskpool is not None:
                    rmsd_mat[i][j] = taskpool.query_task(f'{i}-{j}', block=True, timeout=999)
        return rmsd_mat
    
    def rmsd_cluster(self, n_clusters: int):
        coords = np.concatenate([[pose.parse_coords()] for pose in self.pose_lst], axis=0)
        cluster = KMeans(n_clusters, backend=RmsdClusterBackend())
        return cluster.fit_predict(coords)
    
    def calcu_SSE_SSR(self, rmsd_mat: np.ndarray, groups_idx: np.ndarray,
                      centers_idx: Optional[np.ndarray] = None, center_idx: Optional[int] = None):
        """
        Parameters
            - rmsd_mat: [N_pose, N_pose], RMSD matrix return by DlgFile.rmsd
            - groups_idx: [N_pose], clustering result return by DlgFile.rmsd_cluster or other clustering algorithm
            - centers_idx: [N_cluster], clustering center index, **must in the rmsd matrix**, if not provided, use the minimum of each group
            - center_idx: int, center of the all poses, **must in the rmsd matrix**, if not provided, use the minimum of all poses
            
        Returns
            - SSE: 组内误差平方和
            - SSR: 组间误差平方和
            - SSR/SST: 
        """
        if centers_idx is None:
            centers_idx = np.array([np.argmin(rmsd_mat[groups_idx==i]) for i in np.unique(groups_idx)])
        if center_idx is None:
            center_idx = np.argmin(rmsd_mat)
        sse, ssr = 0, 0
        for i in np.unique(groups_idx):
            sse += np.sum(rmsd_mat[groups_idx==i, centers_idx[i]]**2)
            ssr += np.sum(groups_idx==i) * rmsd_mat[centers_idx[i], center_idx]**2
        return sse, ssr, ssr/(sse+ssr)
    
    def get_top_k_pose(self, method: str = 'energy', k: int = 1, rmsd_mat: np.ndarray = None,
                       groups_idx: np.ndarray = None) -> List[ADModel]:
        if method == 'energy':
            return self.sort_pose(key=lambda x: x.energy, inplace=False)[:k]
        elif method == 'rmsd':
            if rmsd_mat is None:
                rmsd_mat = self.rmsd('numpy')
            if groups_idx is None:
                raise ValueError("groups_idx must be provided if method is 'rmsd'")
            rmsd_vec = rmsd_mat.sum(axis=0)
            idx_vec = np.arange(rmsd_vec.shape[0])
            top_poses = {}
            for group_idx in np.unique(groups_idx):
                sub_idx, sub_rmsd_vec = idx_vec[groups_idx==group_idx], rmsd_vec[groups_idx==group_idx]
                idx = sub_idx[np.argsort(sub_rmsd_vec)[:k]] # increasing order
                top_poses[group_idx] = [self.pose_lst[i] for i in idx]
            return top_poses
        else:
            raise ValueError(f'Unsupported method: {method}')


def merge_dlg(dlg_lst: List[DlgFile]) -> DlgFile:
    if not dlg_lst:
        return put_err('Empty input list, return None')
    new_dlg = DlgFile()
    for dlg in dlg_lst:
        new_dlg.merge(dlg, inplace=True)
    return new_dlg


def tk_file_dialog_wrapper(*args, **kwargs):
    def ret_wrapper(tk_file_dialog_func):
        def core_wrapper(*args, **kwargs):
            parent = tk.Tk()
            result = tk_file_dialog_func(*args, parent = parent, **kwargs)
            parent.withdraw()
            if result == "":
                return None
            else:
                return result
        return core_wrapper
    return ret_wrapper
            

class MyFileDialog:
    def __init__(self, types=[("Executable", "*")], initialdir: str = None):
        self.initialdir = initialdir
        self.types = types

    @tk_file_dialog_wrapper()
    def get_open_file(self, parent):
        return tkFileDialog.askopenfilename(parent=parent, initialdir = self.initialdir, filetypes=self.types)


    @tk_file_dialog_wrapper()
    def get_save_file(self, parent):
        return tkFileDialog.asksaveasfilename(parent=parent, initialdir = self.initialdir, filetypes=self.types)
        
    @tk_file_dialog_wrapper()
    def get_ask_dir(self, parent):
        return tkFileDialog.askdirectory(parent=parent, initialdir = self.initialdir)


if __name__ == '__main__':
    # from mbapy_lite.base import TimeCosts
    # @TimeCosts(10)
    # def load_test(idx):
    #     DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True)
    # load_test()
    normal = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=False, parse2std=False)
    std = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=False, parse2std=True)
    std.parse_energies()
    std.plot_energy_curve()
    plt.show()
    from pymol import cmd
    from rdkit import Chem
    cmd.reinitialize()
    cmd.read_pdbstr(std.pose_lst[0].as_pdb_string(), 'std')
    print(std.pose_lst[0].as_pdb_string(), '\n\n')
    print(cmd.get_pdbstr('std'))
    ligand = Chem.MolFromPDBBlock(cmd.get_pdbstr('std'), removeHs=True, sanitize=False)
    
    import seaborn as sns
    from mbapy_lite.stats import pca
    dlg = DlgFile(path='data_tmp/dlg/1000run.dlg', sort_pdb_line_by_res=True, parse2std=True)
    dlg.sort_pose(inplace=True)
    energies = np.array([pose.energy for pose in dlg.pose_lst])
    rmsd = dlg.rmsd('numpy')
    
    sns.clustermap(rmsd)
    plt.show()
    
    np.random.seed(3407)
    rs = []
    for k in range(2, 7):
        groups_idx = dlg.rmsd_cluster(k)
        sse, ssr, r = dlg.calcu_SSE_SSR(rmsd, groups_idx)
        rs.append(r)
    plt.plot(list(range(2, 7)), rs)
    plt.show()
    
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    group_cmaps = ['Reds', 'Blues', 'Greens']
    groups_idx = dlg.rmsd_cluster(len(group_cmaps))
    top_poses = dlg.get_top_k_pose(method='rmsd', k=1, rmsd_mat=rmsd, groups_idx=groups_idx)
    dots = pca(rmsd, 3)
    for i, group_cmap in enumerate(group_cmaps):
        ax.scatter(dots[groups_idx==i, 0], dots[groups_idx==i, 1], dots[groups_idx==i, 2],
                   c=energies[groups_idx==i], cmap=group_cmap, alpha=0.4, s=40)
    plt.show()
    


'''
Date: 2024-12-13 16:17:09
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-19 18:49:29
Description: 
'''

import os
from pathlib import Path
from typing import Dict, List, Union

from mbapy.base import Configs
from mbapy_lite.web import Browser, random_sleep


def get_login_browser(username: str, password: str, b: Browser = None, download_dir: str = None, timeout: float = 300.0) -> Browser:
    b = b or Browser(download_path=download_dir, use_undetected=True)
    # TODO: set popup blocker
    b.get('https://app.cgenff.com/login')
    b.find_elements('//*[@id="email"]')[0].send_keys(username)
    b.find_elements('//*[@id="password"]')[0].send_keys(password)
    b.find_elements('//*[@id="root"]/main/div/div/div/div/div[5]/button')[0].click()
    b.wait_element(element='//*[@id="root"]/div/main/div/div/div[2]/div[1]/div/p/div/div[3]/div[1]/div/button/img', timeout=timeout)
    return b


def get_result_from_CGenFF(mol2_path: str, username: str = None, password: str = None,
                           b: Browser = None, timeout: float = 30.0) -> Dict[str, Union[str, List[str]]]:
    mol2_path = str(Path(mol2_path).absolute())
    b = b or get_login_broswer(username, password, download_dir=os.path.dirname(mol2_path), timeout=timeout)
    b.get('https://app.cgenff.com/homepage')
    # upload mol2 file and click "Submit" button
    b.click(element='//*[@id="root"]/div/main/div/div/div[2]/div[1]/div/p/div/div[3]/div[1]/div/button')
    upload = b.find_elements('/html/body/div[2]/div[3]/div/div[1]/div/input')[0]
    upload.send_keys(mol2_path)
    b.click(element='/html/body/div[2]/div[3]/div/div[2]/button[2]')
    # check and run
    b.click(element='//*[@id="root"]/div/main/div/div[1]/div/div[3]/div[2]/button[2]')
    b.click(element='//*[@id="root"]/div/main/div/div[2]/button[1]')
    # download result
    b.click(element='//*[@id="root"]/div/main/div/div[2]/button[2]')
    # run generate gmx result and wait, click "Download" button
    b.click(element='//*[@id="root"]/div/main/div/div[2]/button[3]')
    while b.find_elements('//*[@id="root"]/div/main/div/div[2]/button[3]')[0].text != 'DOWNLOAD GROMACS FORMAT':
        random_sleep(2)
    b.click(element='//*[@id="root"]/div/main/div/div[2]/button[3]')


if __name__ == '__main__':
    from mbapy.file import opts_file
    config = opts_file('data_tmp/config.json', way='json')
    mol2_path = 'data_tmp/pdb/ligand.mol2'
    mol2_path = str(Path(mol2_path).absolute())
    b = Browser(options = [f"--user-agent={Configs.web.chrome_driver_path}", "--disable-popup-blocking"], download_path=os.path.dirname(mol2_path), use_undetected=True)
    b = get_login_browser(config['CGENFF']['username'], config['CGENFF']['password'], b=b, download_dir=os.path.dirname(mol2_path))
    get_result_from_CGenFF(mol2_path, b=b)


import itertools
from typing import List, Union

import numpy as np
from compas.geometry import oriented_bounding_box_numpy
from mbapy.base import put_err, split_list
from pymol import cmd, editing


def sort_vertices(vertices, sort_orders=[(0, 1, 2), (1, 2, 0), (2, 0, 1)], tolerance=0.1):
    """
    对顶点进行排序，以匹配plot_bounding_box中的顺序。
    
    参数:
    vertices (numpy.ndarray): 8个顶点的坐标，形状为 (8, 3)。
    sort_orders (list of tuples): 排序顺序的列表，每个元组包含三个索引，表示XYZ的排序顺序。
    
    返回:
    list: [P0, P0-x, ...] 或 None（如果所有排序顺序都失败）。
    """
    neighbors = {
        0: [1, 2, 4], 1: [0, 3, 5], 2: [0, 3, 6], 3: [1, 2, 7], 4: [0, 5, 6], 5: [1, 4, 7], 6: [2, 4, 7], 7: [3, 5, 6]
    }
    def is_orthogonal(v1, v2, tolerance=1e-6):
        """检查两个向量是否正交，包含指定的误差容许"""
        _v1, _v2 = v1 / np.linalg.norm(v1), v2 / np.linalg.norm(v2)
        return abs(np.dot(_v1, _v2)) < tolerance

    def check_orthogonality(points, tolerance=1e-6):
        """检查所有顶点的三条棱是否正交"""
        for i in range(8):
            vectors = []
            for j in neighbors[i]:
                vectors.append(np.array(points[j]) - np.array(points[i]))
            for v1, v2 in itertools.combinations(vectors, 2):
                if not is_orthogonal(v1, v2, tolerance):
                    return False
        return True

    for sort_order in sort_orders:
        vertices_copy = vertices.copy()
        for i, j in zip(sort_order, [4, 2, 1]):
            tmp = []
            for sub in split_list(vertices_copy, j):
                tmp.extend(sorted(sub, key=lambda v: v[i]))
            vertices_copy = tmp
        
        if check_orthogonality(vertices_copy, tolerance):
            return vertices_copy
    
    raise ValueError("No valid sorting order found for the given vertices.")


def calcu_bounding_box(pml_name: str = None, coords: np.ndarray = None, state: int = 0):
    index2coords = {}
    if coords is None:
        cmd.iterate_state(state, pml_name, 'index2coords[index] = [x, y, z]', space=locals())
        coords = np.array(list(index2coords.values()))
        index2coords = {index: i for i, index in enumerate(list(index2coords.keys()))}
    if not coords.any():
        return put_err('No coordinates found in selection, return None.', None)
    bounding_box_vertices = oriented_bounding_box_numpy(coords)
    sorted_vertices = sort_vertices(bounding_box_vertices)
    return coords, index2coords, sorted_vertices


def align_bounding_box_to_axis(coords: np.ndarray, bounding_box_vertices: np.ndarray,
                               fixed_coords: Union[List[float], str] = 'center'):
    """
    以指定的点为不动点对点云进行旋转，使其包围盒的各边与坐标轴对齐。
    
    参数:
    coords (numpy.ndarray): 点云数据，形状为 (N, 3)。
    fixed_coords (numpy.ndarray): 不动点坐标，形状为 (3,)。
    
    返回:
    numpy.ndarray: 旋转后的点云数据。
    """
    if isinstance(fixed_coords, str):
        if fixed_coords == 'center':
            fixed_coords = coords.mean(axis=0)
        else:
            return put_err(f'Unsupported fixed_point value: {fixed_coords}, return original vertices.', bounding_box_vertices)
    bounding_box_vertices = np.array(bounding_box_vertices)
    # Select the first vertex and its three adjacent vertices
    v0 = bounding_box_vertices[0]
    v1 = bounding_box_vertices[1]
    v2 = bounding_box_vertices[2]
    v4 = bounding_box_vertices[4]
    # Compute the three principal axes from the bounding box edges
    axis_x = v1 - v0
    axis_y = v2 - v0
    axis_z = v4 - v0
    # Normalize the axes to get unit vectors
    axis_x = axis_x / np.linalg.norm(axis_x)
    axis_y = axis_y / np.linalg.norm(axis_y)
    axis_z = axis_z / np.linalg.norm(axis_z)
    # Construct the rotation matrix from the principal axes
    rotation_matrix = np.column_stack((axis_x, axis_y, axis_z))
    # Rotate the point cloud around the fixed point
    rotated_points = np.dot(coords - fixed_coords, rotation_matrix) + fixed_coords
    rotated_box = np.dot(bounding_box_vertices - fixed_coords, rotation_matrix) + fixed_coords
    return rotated_points, rotated_box, rotation_matrix, fixed_coords


def normalize(v):
    """归一化向量"""
    norm = np.linalg.norm(v)
    if norm == 0:
        return v
    return v / norm

def angle_between(v1, v2):
    """计算两个向量之间的夹角（弧度）"""
    cos_theta = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
    return np.arccos(np.clip(cos_theta, -1.0, 1.0))

def rotation_matrix_from_vectors(v1, v2):
    """计算从向量v1旋转到向量v2的旋转矩阵"""
    v1 = normalize(v1)
    v2 = normalize(v2)
    v = np.cross(v1, v2)
    s = np.linalg.norm(v)
    c = np.dot(v1, v2)
    vx = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    if s == 0:
        # 如果v1和v2同向或反向，不需要旋转
        return np.eye(3)
    R = np.eye(3) + vx + np.dot(vx, vx) * ((1 - c) / (s ** 2))
    return R

def rotation_matrix_x(angle):
    """绕X轴旋转的旋转矩阵"""
    c = np.cos(np.radians(angle))
    s = np.sin(np.radians(angle))
    return np.array([[1, 0, 0],
                     [0, c, -s],
                     [0, s, c]])

def rotation_matrix_y(angle):
    """绕Y轴旋转的旋转矩阵"""
    c = np.cos(np.radians(angle))
    s = np.sin(np.radians(angle))
    return np.array([[c, 0, s],
                     [0, 1, 0],
                     [-s, 0, c]])

def rotation_matrix_z(angle):
    """绕Z轴旋转的旋转矩阵"""
    c = np.cos(np.radians(angle))
    s = np.sin(np.radians(angle))
    return np.array([[c, -s, 0],
                     [s, c, 0],
                     [0, 0, 1]])

def apply_rotation(coords, angles):
    """应用旋转到顶点"""
    # 构建旋转矩阵
    Rx = rotation_matrix_x(angles[0])
    Ry = rotation_matrix_y(angles[1])
    Rz = rotation_matrix_z(angles[2])
    
    # 组合旋转矩阵
    R = np.dot(Rz, np.dot(Ry, Rx))
    
    # 应用旋转矩阵到每个顶点
    rotated_vertices = np.dot(coords, R.T)
    return rotated_vertices

def align_edge_with_x_axis(v0, v1):
    """计算将边v0到v1与X轴对齐所需的旋转角度"""
    edge1 = np.array(v1) - np.array(v0)
    x_axis = np.array([1, 0, 0])
    R = rotation_matrix_from_vectors(edge1, x_axis)
    # 计算欧拉角
    sy = np.sqrt(R[0, 0] * R[0, 0] + R[1, 0] * R[1, 0])
    singular = sy < 1e-6
    if not singular:
        x = np.arctan2(R[2, 1], R[2, 2])
        y = np.arctan2(-R[2, 0], sy)
        z = np.arctan2(R[1, 0], R[0, 0])
    else:
        x = np.arctan2(-R[1, 2], R[1, 1])
        y = np.arctan2(-R[2, 0], sy)
        z = 0
    # 将弧度转换为度
    euler_angles = np.degrees([x, y, z])
    return euler_angles

def align_edge_with_y_axis_after_x_alignment(v0, v2):
    """计算在v0-v1与X轴对齐后，将v0-v2与Y轴对齐所需的绕X轴旋转角度"""
    edge2 = np.array(v2) - np.array(v0)
    # 只考虑Y和Z分量，因为X分量在与X轴对齐后为0
    yz_projection = edge2[1:]  # 取Y和Z分量
    y_axis = np.array([1, 0])  # Y轴在YZ平面上的投影
    angle = np.arctan2(yz_projection[1], yz_projection[0])  # 计算绕X轴旋转角度
    return np.degrees(angle)

def rotate_bounding_box_to_axis(bounding_box_vertices: np.ndarray):
    """
    计算包围盒绕轴旋转的角度，使包围盒的各边与坐标轴对齐。
    
    参数:
    bounding_box_vertices： 包围盒的8个顶点坐标，形状为 (8, 3)。
    
    返回:
    numpy.ndarray: 旋转角度。
    """
    bounding_box_vertices = np.array(bounding_box_vertices)
    angles1 = align_edge_with_x_axis(bounding_box_vertices[0], bounding_box_vertices[1])
    vertices_rotated = apply_rotation(bounding_box_vertices, angles1)
    angle2 = align_edge_with_y_axis_after_x_alignment(vertices_rotated[0], vertices_rotated[2])
    return angles1.tolist(), [angle2, 0, 0]


def align_pose_to_axis(pml_name: str, move_name: str = None, fixed: Union[List[float], str] = 'center', state: int = 0,
                       move_method: str = 'rotate', dss: bool = True, quite: int = 1):
    """
    Parameters:
        - pml_name (str): pymol object name to calculate minimum bounding box and tranformation matrix.
        - move_name (str): pymol object name to move to aligned position.
        - fixed (Union[List[float], str]): fixed point to align with, not works when move_method is 'rotate'.
        - state (int): pml state to calculate minimum bounding box.
        - move_method (str): method to move to aligned position, support 'transform', 'alter' and 'rotate'.
        - dss (bool): whether to apply dss to pml object.
        - quite (int): whether to print log information.
    
    Returns:
        - aligned_coords (numpy.ndarray): aligned coordinates.
        - aligned_box (numpy.ndarray): aligned bounding box.
        - rotation_matrix (numpy.ndarray): rotation matrix to align with.
        - fixed_coords (numpy.ndarray): fixed point to align with.

    Notes:
        - TODO: RMS is rigth(=0), but second structure all trun to raandom coil, fixed by rotate method.
        - TODO: in rotate method, apply angles1 gets right result, but apply angles2 gets wrong result, pymol is fine.
    """
    move_name = move_name or pml_name
    # get coords
    coords, index2coords, sorted_vertices = calcu_bounding_box(pml_name, state=state)
    if isinstance(fixed, str) and fixed != 'center':
        fixed_coords = []
        cmd.iterate_state(state, fixed, 'fixed_coords.append([x, y, z])', space=locals())
        fixed_coords = np.array(fixed_coords).mean(axis=0)
    else:
        fixed_coords = fixed
    if move_method in {'transform', 'alter'}:
        # align bounding box
        aligned_coords, aligned_box, rotation_matrix, fixed_coords = align_bounding_box_to_axis(coords, sorted_vertices, fixed_coords=fixed_coords)
        # create pymol rotation matrix
        pml_mat = np.zeros((4, 4))
        pml_mat[:3, :3] = rotation_matrix.T # np is matmul, but pymol is dot product
        pml_mat[-1, :] = list(-fixed_coords) + [1]
        pml_mat[:, -1] = list(fixed_coords) + [1]
        if not quite:
            print(f'pymol transform matrix: {pml_mat.flatten().tolist()}')
    elif move_method == 'rotate':
        # calculate rotation degree
        angles1, angles2 = rotate_bounding_box_to_axis(sorted_vertices)
        if not quite:
            print(f'Rotate angles1: {angles1}, angles2: {angles2}')
    else:
        return put_err(f'Unsupported move_method: {move_method}, only support transform and alter, skip transform.')
    # move to aligned position
    if move_method == 'transform':
        cmd.transform_selection(move_name, pml_mat.flatten().tolist(), homogenous=0)
    elif move_method == 'alter':
        cmd.alter_state(state, move_name, 'x = aligned_coords[index2coords[index], 0]', space=locals())
        cmd.alter_state(state, move_name, 'y = aligned_coords[index2coords[index], 1]', space=locals())
        cmd.alter_state(state, move_name, 'z = aligned_coords[index2coords[index], 2]', space=locals())
        cmd.sort(move_name)
    elif move_method == 'rotate':
        for angles in [angles1, angles2]:
            for axis, angle in zip(['x', 'y', 'z'], angles):
                if angle != 0:
                    cmd.rotate(axis, angle, move_name)
            aligned_coords = apply_rotation(coords, angles)
            aligned_box = apply_rotation(sorted_vertices, angles)
        rotation_matrix = None
    if dss:
        editing.dss(move_name)
    cmd.rebuild(move_name)
    return aligned_coords, aligned_box, rotation_matrix, fixed_coords

def set_axes_equal(ax):
    """
    设置坐标轴的比例相等。
    
    参数:
    ax (Axes3D): 3D坐标轴。
    """
    xlim = ax.get_xlim3d()
    ylim = ax.get_ylim3d()
    zlim = ax.get_zlim3d()

    x_range = abs(xlim[1] - xlim[0])
    x_middle = np.mean(xlim)
    y_range = abs(ylim[1] - ylim[0])
    y_middle = np.mean(ylim)
    z_range = abs(zlim[1] - zlim[0])
    z_middle = np.mean(zlim)

    # 找到最大的范围
    plot_radius = 0.5 * max([x_range, y_range, z_range])

    ax.set_xlim3d([x_middle - plot_radius, x_middle + plot_radius])
    ax.set_ylim3d([y_middle - plot_radius, y_middle + plot_radius])
    ax.set_zlim3d([z_middle - plot_radius, z_middle + plot_radius])


def plot_bounding_box(ax, vertices, color='r'):
    """
    绘制长方体的边。
    
    参数:
    ax (Axes3D): 3D坐标轴。
    vertices (list): 长方体的8个顶点坐标。
    """
    # 定义长方体的12条边
    edges = [
        (0, 1), (1, 3), (3, 2), (2, 0),  # 底部面
        (4, 5), (5, 7), (7, 6), (6, 4),  # 顶部面
        (0, 4), (1, 5), (2, 6), (3, 7)   # 连接上下两个面
    ]
    
    # 绘制每条边
    for edge in edges:
        ax.plot(*zip(vertices[edge[0]], vertices[edge[1]]), linestyle='-', color=color)
    # 标记每个点
    for i in range(8):
        ax.text(vertices[i][0], vertices[i][1], vertices[i][2], str(i))


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    from lazydock.pml.thirdparty.draw_bounding_box import draw_bounding_box
    
    cmd.reinitialize()
    cmd.load('data_tmp/pdb/RECEPTOR.pdb', 'receptor')
    cmd.select('receptor_CA', 'name CA and receptor')
    coords, _, vertics = calcu_bounding_box('receptor_CA')
    aligned_coords, aligned_box, rotation_matrix, fixed_coords = align_bounding_box_to_axis(coords, vertics)
    _, _, aligned_vertics = calcu_bounding_box(coords=aligned_coords)
    
    aligned_coords, aligned_box, rotation_matrix, fixed_coords = align_pose_to_axis('receptor', move_method='rotate')
    draw_bounding_box('receptor')
    cmd.save('data_tmp/pdb/RECEPTOR_bounding_box.pse')
    
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(aligned_coords[:, 0], aligned_coords[:, 1], aligned_coords[:, 2], c='r', alpha=0.5, marker='o', s=10, label='Original Points')
    plot_bounding_box(ax, aligned_box, color='r')
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    set_axes_equal(ax)
    ax.legend()
    plt.show()
    


'''
Date: 2024-12-21 08:49:55
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-26 10:38:56
Description: steps most from http://www.mdtutorials.com/gmx
'''


import argparse
import os
import re
import shutil
from pathlib import Path
from typing import Dict, List, Tuple, Union

from mbapy_lite.base import Configs, put_err, put_log
from mbapy_lite.file import get_paths_with_extension, opts_file
from pymol import cmd
from tqdm import tqdm

from lazydock.gmx.run import Gromacs
from lazydock.scripts._script_utils_ import Command, clean_path
from lazydock.utils import uuid4


class simple_protein(Command):
    HELP = """
    run simple protein GROMACS simulation
    1. gmx editconf -f protein.gro -o protein_newbox.gro -c -d 1.0 -bt cubic
    2. gmx solvate -cp protein_newbox.gro -cs spc216.gro -o protein_solv.gro -p topol.top
    3. gmx grompp -f ions.mdp -c protein_solv.gro -p topol.top -o ions.tpr
    4. gmx genion -s ions.tpr -o protein_solv_ions.gro -p topol.top -pname NA -nname CL -neutral
    5. gmx grompp -f minim.mdp -c protein_solv_ions.gro -p topol.top -o em.tpr
    6. gmx mdrun -v -deffnm em
    7. gmx energy -f em.edr -o potential.xvg # At the prompt, type "10 0" to select Potential (10); zero (0) terminates input.
    8. gmx grompp -f nvt.mdp -c em.gro -r em.gro -p topol.top -o nvt.tpr
    9. gmx mdrun -deffnm nvt
    10. gmx energy -f nvt.edr -o temperature.xvg # Type "16 0" at the prompt to select the temperature of the system and exit.
    11. gmx grompp -f npt.mdp -c nvt.gro -r nvt.gro -t nvt.cpt -p topol.top -o npt.tpr
    12. gmx mdrun -deffnm npt
    13. gmx energy -f npt.edr -o pressure.xvg # Type "18 0" at the prompt to select the pressure of the system and exit.
    14. gmx energy -f npt.edr -o density.xvg # using energy and entering "24 0" at the prompt.
    15. gmx grompp -f md.mdp -c npt.gro -t npt.cpt -p topol.top -o md.tpr
    16. gmx mdrun -v -ntomp 4 -deffnm md -update gpu -nb gpu -pme gpu -bonded gpu -pmefft gpu
    """
    def __init__(self, args, printf=print):
        super().__init__(args, printf)
        self.indexs = {}
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args.add_argument('-d', '--dir', type=str,
                          help='directory to store the prepared files')
        args.add_argument('-n', '--protein-name', type = str,
                          help='protein name in each sub-directory, such as protein.gro.')
        args.add_argument('--auto-box', action='store_true', default=False,
                          help='FLAG, whether to automatically generate rectangular bounding box via on pymol.cmd.get_extent.')
        args.add_argument('--auto-box-padding', type=float, default=1.2,
                          help='distance, padding the box size, default is %(default)s.')
        args.add_argument('--ion-mdp', type = str,
                          help='energy minimization mdp file, if is a file-path, will copy to working directory; if is a file-name, will search in working directory.')
        args.add_argument('--em-mdp', type = str,
                          help='energy minimization mdp file, if is a file-path, will copy to working directory; if is a file-name, will search in working directory.')
        args.add_argument('--nvt-mdp', type = str,
                          help='nvt mdp file, if is a file-path, will copy to working directory; if is a file-name, will search in working directory.')
        args.add_argument('--npt-mdp', type = str,
                          help='npt mdp file, if is a file-path, will copy to working directory; if is a file-name, will search in working directory.')
        args.add_argument('--md-mdp', type = str,
                          help='production md mdp file, if is a file-path, will copy to working directory; if is a file-name, will search in working directory.')
        args.add_argument('--editconf-args', type = str, default="-c -d 1.2 -bt dodecahedron",
                          help='args pass to editconf command, default is %(default)s.')
        args.add_argument('--solvate-args', type = str, default="-cs spc216.gro",
                          help='args pass to solvate command, default is %(default)s.')
        args.add_argument('--genion-args', type = str, default="-pname NA -nname CL -neutral",
                          help='args pass to genion command, default is %(default)s.')
        args.add_argument('--mdrun-args', type = str, default="-v -ntomp 4 -update gpu -nb gpu -pme gpu -bonded gpu -pmefft gpu",
                          help='args pass to mdrun command, default is %(default)s.')                          
        return args
    
    def process_args(self):
        self.args.dir = clean_path(self.args.dir)
        
    def get_mdp(self, working_dir: Path):
        mdps = {}
        for name in ['ion', 'em', 'nvt', 'npt','md']:
            mdp_file = getattr(self.args, f'{name}_mdp')
            if os.path.isfile(mdp_file):
                shutil.copy(mdp_file, working_dir)
                mdps[name] = Path(mdp_file).name
            elif os.path.isfile(working_dir / mdp_file):
                mdps[name] = (working_dir / mdp_file).name
            else:
                put_err(f'can not find {name} mdp file: {mdp_file} in {mdp_file} or {working_dir}, exit.', _exit=True)
        return mdps
    
    @staticmethod
    def get_box(mol_path: Path, padding: float, sele: str = 'not resn SOL'):
        cmd.reinitialize()
        name = uuid4()
        cmd.load(str(mol_path), mol_path.name)
        cmd.select(name, sele)
        ([minX, minY, minZ], [maxX, maxY, maxZ]) = cmd.get_extent(name)
        box_center = list(map(lambda x: x/20, [maxX+minX, maxY+minY, maxZ+minZ]))
        box_size = list(map(lambda x: x/10+2*padding, [maxX-minX, maxY-minY, maxZ-minZ]))
        cmd.reinitialize()
        return box_center, box_size
    
    def make_box(self, protein_path: Path, main_name: str, gmx: Gromacs, mdps: Dict[str, str]):
        # STEP 1: editconf -f protein.gro -o protein_newbox.gro -c -d 1.0 -bt cubic
        if self.args.auto_box:
            # get shift from first editconf
            _, box_size = self.get_box(protein_path, self.args.auto_box_padding)
            manual_box_cmd = f'-box {" ".join(map(lambda x: f"{x:.2f}", box_size))}'
            editconf_args = self.args.editconf_args.replace('-d 1.2 -bt dodecahedron', ' ') + manual_box_cmd
            _, log_path = gmx.run_command_with_expect(f'editconf {editconf_args}', f=f'{main_name}.gro', o=f'{main_name}_newbox_tmp.gro', enable_log=True)
            shift_line = list(filter(lambda x: 'new center' in x.strip(), opts_file(log_path, way='lines')))[0]
            shift = list(map(float, re.findall(r'[\d\-\.]+', shift_line)))
            # get solvated box from first solvate
            shutil.copy(protein_path.parent / 'topol.top', protein_path.parent / 'topol_tmp.top')
            gmx.run_command_with_expect(f'solvate {self.args.solvate_args}', cp=f'{main_name}_newbox_tmp.gro', o=f'{main_name}_solv_tmp.gro', p='topol_tmp.top')
            solv_center, solv_size = self.get_box(protein_path.parent / f'{main_name}_solv_tmp.gro', self.args.auto_box_padding, 'resn SOL')
            prot_center, _ = self.get_box(protein_path.parent / f'{main_name}_newbox_tmp.gro', self.args.auto_box_padding)
            put_log(f'protein box size: {box_size}, tmp solvated box size: {solv_size}, protein center: {prot_center}, tmp solvated center: {solv_center}, shift: {shift}')
            # calculate new box center
            box_center = [s+(x1-x2) for s, x1, x2 in zip(shift, solv_center, prot_center)]
            # run editconf with new box center and size
            editconf_args += f' -center {" ".join(map(lambda x: f"{x:.2f}", box_center))}'
            _, log_path = gmx.run_command_with_expect(f'editconf {editconf_args}', f=f'{main_name}.gro', o=f'{main_name}_newbox.gro', enable_log=True)
        else:
            gmx.run_command_with_expect(f'editconf {self.args.editconf_args}', f=f'{main_name}.gro', o=f'{main_name}_newbox.gro')
        # STEP 2: solvate -cp protein_newbox.gro -cs spc216.gro -o protein_solv.gro -p topol.top
        gmx.run_command_with_expect(f'solvate {self.args.solvate_args}', cp=f'{main_name}_newbox.gro', o=f'{main_name}_solv.gro', p='topol.top')
        # STEP 3: grompp -f ions.mdp -c protein_solv.gro -p topol.top -o ions.tpr
        gmx.run_command_with_expect('grompp', f=mdps['ion'], c=f'{main_name}_solv.gro', p='topol.top', o='ions.tpr')
        # STEP 4: genion -s ions.tpr -o protein_solv_ions.gro -p topol.top -pname NA -nname CL -neutral
        gmx.run_command_with_expect(f'genion {self.args.genion_args}', s='ions.tpr', o=f'{main_name}_solv_ions.gro', p='topol.top',
                                    expect_actions=[{'Select a group:': '13\r', 'No ions to add': '', '\\timeout': ''}],
                                    expect_settings={'start_timeout': 600})
        
    def energy_minimization(self, protein_path: Path, main_name: str, gmx: Gromacs, mdps: Dict[str, str]):
        # STEP 5: grompp -f minim.mdp -c protein_solv_ions.gro -p topol.top -o em.tpr
        gmx.run_command_with_expect('grompp', f=mdps['em'], c=f'{main_name}_solv_ions.gro', p='topol.top', o='em.tpr')
        # STEP 6: mdrun -v -deffnm em
        gmx.run_command_with_expect(f'mdrun', deffnm='em')
        # STEP 7: energy -f em.edr -o potential.xvg
        gmx.run_command_with_expect('energy', f='em.edr', o='potential.xvg',
                                    expect_actions=[{'T-rest': '11 0\r'}])
        os.system(f'cd "{protein_path.parent}" && dit xvg_show -f potential.xvg -o potential.png -smv')
        
    def equilibration(self, protein_path: Path, main_name: str, gmx: Gromacs, mdps: Dict[str, str]):
        # STEP 8: grompp -f nvt.mdp -c em.gro -r em.gro -p topol.top -o nvt.tpr
        gmx.run_command_with_expect('grompp', f=mdps['nvt'], c='em.gro', r='em.gro', p='topol.top', o='nvt.tpr', n=self.indexs.get('nvt', None))
        # STEP 9: mdrun -deffnm nvt
        gmx.run_command_with_expect('mdrun', deffnm='nvt')
        # STEP 10: energy -f nvt.edr -o temperature.xvg
        gmx.run_command_with_expect('energy', f='nvt.edr', o='temperature.xvg',
                                    expect_actions=[{'Lamb-non-Protein': '16 0\r'}])
        os.system(f'cd "{protein_path.parent}" && dit xvg_show -f temperature.xvg -o temperature.png -smv')
        # STEP 11: grompp -f npt.mdp -c nvt.gro -r nvt.gro -t nvt.cpt -p topol.top -o npt.tpr
        gmx.run_command_with_expect('grompp', f=mdps['npt'], c='nvt.gro', r='nvt.gro', t='nvt.cpt', p='topol.top', o='npt.tpr', n=self.indexs.get('npt', None))
        # STEP 12: mdrun -deffnm npt
        gmx.run_command_with_expect('mdrun', deffnm='npt')
        # STEP 13: energy -f npt.edr -o pressure.xvg
        gmx.run_command_with_expect('energy', f='npt.edr', o='pressure.xvg',
                                    expect_actions=[{'Lamb-non-Protein': '17 0\r'}])
        os.system(f'cd "{protein_path.parent}" && dit xvg_show -f pressure.xvg -o pressure.png -smv')
        # STEP 14: energy -f npt.edr -o density.xvg
        gmx.run_command_with_expect('energy', f='npt.edr', o='density.xvg',
                                    expect_actions=[{'Lamb-non-Protein': '23 0\r'}])
        os.system(f'cd "{protein_path.parent}" && dit xvg_show -f density.xvg -o density.png -smv')
        
    def production_md(self, protein_path: Path, main_name: str, gmx: Gromacs, mdps: Dict[str, str]):
        # STEP 15: grompp -f md.mdp -c npt.gro -t npt.cpt -p topol.top -o md.tpr
        gmx.run_command_with_expect('grompp', f=mdps['md'], c='npt.gro', t='npt.cpt', p='topol.top', o='md.tpr', n=self.indexs.get('md', None))
        # STEP 16: mdrun -v -ntomp 4 -deffnm md -update gpu -nb gpu -pme gpu -bonded gpu -pmefft gpu
        gmx.run_command_with_expect(f'mdrun {self.args.mdrun_args}', deffnm='md')
        
    def main_process(self):
        # get protein paths
        if os.path.isdir(self.args.dir):
            proteins_path = get_paths_with_extension(self.args.dir, [], name_substr=self.args.protein_name)
        else:
            put_err(f'dir argument should be a directory: {self.args.dir}, exit.', _exit=True)
        put_log(f'get {len(proteins_path)} protein(s)')
        # process each complex
        for protein_path in tqdm(proteins_path, total=len(proteins_path)):
            protein_path = Path(protein_path).resolve()
            main_name = protein_path.stem
            gmx = Gromacs(working_dir=str(protein_path.parent))
            mdps = self.get_mdp(protein_path.parent)
            # STEP 1 ~ 4: make box, solvate, ions
            self.make_box(protein_path, main_name, gmx, mdps)
            # STEP 5 ~ 7: energy minimization
            self.energy_minimization(protein_path, main_name, gmx, mdps)
            # STEP 8 ~ 14: equilibration
            self.equilibration(protein_path, main_name, gmx, mdps)
            # STEP 15 ~ 16: production md
            self.production_md(protein_path, main_name, gmx, mdps)


class simple_ligand(simple_protein):
    HELP = simple_protein.HELP.replace('protein', 'ligand')
    def __init__(self, args, printf=print):
        super().__init__(args, printf)
            
    
class simple_complex(simple_ligand):
    HELP = simple_protein.HELP.replace('protein', 'complex')
    def __init__(self, args, printf=print):
        super().__init__(args, printf)
        
    @staticmethod
    def make_args(args: argparse.ArgumentParser):
        args = simple_protein.make_args(args)
        args.add_argument('-ln', '--ligand-name', type = str, default='lig.gro',
                          help='ligand name in each sub-directory, such as lig.gro, default is %(default)s.')
        args.add_argument('--lig-posres', type=str, default='POSRES',
                          help='ligand position restraint symbol, default is %(default)s.')
        args.add_argument('--tc-groups', type=str, default='1 | 13',
                          help='tc-grps to select, so could set tc-grps = Protein_JZ4 Water_and_ions to achieve "Protein Non-Protein" effect., default is %(default)s.')
        return args
        
    def equilibration(self, protein_path: Path, main_name: str, gmx: Gromacs, mdps: Dict[str, str]):
        from lazydock.scripts.prepare_gmx import ligand
        lig_name = Path(self.args.ligand_name).stem
        # STEP 1: make restraints file for ligand
        # make_ndx -f jz4.gro -o index_jz4.ndx
        gmx.run_command_with_expect('make_ndx', f=f'{lig_name}.gro', o=f'index_{lig_name}.ndx',
                                    expect_actions=[{'>': '0 & ! a H*\r'}, {'>': 'q\r'}])
        # gmx genrestr -f jz4.gro -n index_jz4.ndx -o posre_jz4.itp -fc 1000 1000 1000
        gmx.run_command_with_expect('genrestr', f=f'{lig_name}.gro', n=f'index_{lig_name}.ndx', o=f'posre_{lig_name}.itp', fc='1000 1000 1000')
        # STEP 2: add restraints info into topol.top
        res_info = f'\n; Ligand position restraints\n#ifdef {self.args.lig_posres}\n#include "posre_{lig_name}.itp"\n#endif\n\n'
        ligand.insert_content(protein_path.parent / 'topol.top', f'#include "{lig_name}.itp"\n', res_info)
        # STEP 3: make tc-grps index file
        # gmx make_ndx -f em.gro -o index.ndx
        gmx.run_command_with_expect('make_ndx', f='em.gro', o='tc_index.ndx',
                                    expect_actions=[{'>': f'{self.args.tc_groups}\r'}, {'>': 'q\r'}])
        for k in ['nvt', 'npt','md']:
            self.indexs[k] = 'tc_index.ndx'


_str2func = {
    'simple-protein': simple_protein,
    'simple-ligand': simple_ligand,
    'simple-complex': simple_complex,
}

def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'tools for GROMACS.')
    subparsers = args_paser.add_subparsers(title='subcommands', dest='sub_command')

    simple_protein_args = simple_protein.make_args(subparsers.add_parser('simple-protein', description=simple_protein.HELP))
    simple_ligand_args = simple_ligand.make_args(subparsers.add_parser('simple-ligand', description=simple_ligand.HELP))
    simple_complex_args = simple_complex.make_args(subparsers.add_parser('simple-complex', description=simple_complex.HELP))

    args = args_paser.parse_args(sys_args)
    if args.sub_command in _str2func:
        _str2func[args.sub_command](args).excute()


if __name__ == "__main__":
    # pass
    # main(r'complex -d data_tmp/gmx/complex -n complex.pdb --receptor-chain-name A --ligand-chain-name Z --ff-dir data_tmp/gmx/charmm36-jul2022.ff'.split())
    
    main()


'''
Date: 2024-12-07 20:24:12
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-12-10 10:09:30
Description: 
'''
import os
from pathlib import Path
from typing import Dict, List, Union

import pandas as pd
import requests
from mbapy_lite.web import Browser, download_streamly, random_sleep


def gridbox2residues(gridbox: str) -> List[str]:
    pass


def run_dock_on_HDOCK(receptor_path: str, ligand_path: str, w_dir: str = None,
                      email: str = None, browser: Browser = None) -> Dict[str, Union[str, float]]:
    """"""
    receptor_path, ligand_path = str(Path(receptor_path).resolve()), str(Path(ligand_path).resolve())
    w_dir = w_dir or Path(receptor_path).resolve().parent
    b = browser or Browser()
    b.browser.set_page_load_timeout(10000)
    b.get('http://hdock.phys.hust.edu.cn/')
    # upload receptor and ligand pdb files
    for path, xpath in zip([receptor_path, ligand_path], ['//*[@id="pdbfile1"]', '//*[@id="pdbfile2"]']):
        btn = b.find_elements(xpath)[0]
        btn.send_keys(path)
    # set email
    if email is not None:
        b.send_keys('//*[@id="emailaddress"]', email)
    # submit and wait for result
    b.click(element='//*[@id="form1"]/table/tbody/tr[9]/td/input[1]')
    xpath_result_url = '/html/body/center/table[1]/tbody/tr/td/div/a[25]'
    while not b.find_elements(xpath_result_url):
        random_sleep(100)
        b.browser.refresh()
    # download result
    result_url = b.browser.current_url + 'all_results.tar.gz'
    result_path = os.path.join(w_dir, 'HDOCK_all_results.tar.gz')
    download_streamly(result_url, result_path, requests.Session())
    # parse score, if 'too many atoms in the lig residue UNK', result file will not contain score
    tabel = b.find_elements('/html/body/center/table[3]/tbody')[0].text
    scores = list(map(lambda x: x.split(), tabel.split('\n')))
    df_data = list(map(list, zip(*[s[-10:] for s in scores])))
    df = pd.DataFrame(df_data, columns=[' '.join(s[:-10]) for s in scores])
    df.to_excel(os.path.join(w_dir, 'HDOCK_score.xlsx'), index=False)
    return {'Score': df}


def run_dock_on_HPEPDOCK(receptor_path: str, ligand_path: str, w_dir: str = None,
                         email: str = None, browser: Browser = None) -> Dict[str, Union[str, float]]:
    """"""
    receptor_path, ligand_path = str(Path(receptor_path).resolve()), str(Path(ligand_path).resolve())
    w_dir = w_dir or Path(receptor_path).resolve().parent
    b = browser or Browser()
    b.browser.set_page_load_timeout(10000)
    b.get('http://huanglab.phys.hust.edu.cn/hpepdock/')
    # upload receptor and ligand pdb files
    for path, xpath in zip([receptor_path, ligand_path], ['//*[@id="pdbfile1"]', '//*[@id="pdbfile2"]']):
        btn = b.find_elements(xpath)[0]
        btn.send_keys(path)
    # set email
    if email is not None:
        b.send_keys('//*[@id="emailaddress"]', email)
    # submit and wait for result
    b.click(element='//*[@id="form1"]/table/tbody/tr[8]/td/input[1]')
    xpath_result_url = '/html/body/center/table[1]/tbody/tr/td/div/a[24]'
    while not b.find_elements(xpath_result_url):
        random_sleep(100)
        b.browser.refresh()
    # download result
    result_url = b.browser.current_url + 'all_results.tar.gz'
    result_path = os.path.join(w_dir, 'HPEPDOCK_all_results.tar.gz')
    download_streamly(result_url, result_path, requests.Session())
    # parse score, if 'too many atoms in the lig residue UNK', result file will not contain score
    tabel = b.find_elements('/html/body/center/table[3]/tbody')[0].text
    scores = list(map(lambda x: x.split(), tabel.split('\n')))
    df_data = list(map(list, zip(*[s[-10:] for s in scores])))
    df = pd.DataFrame(df_data, columns=[' '.join(s[:-10]) for s in scores])
    df.to_excel(os.path.join(w_dir, 'HPEPDOCK_score.xlsx'), index=False)
    return {'Score': df}


if __name__ == '__main__':
    receptor_path = 'data_tmp/docking/ligand1/receptor.pdb'
    ligand_path = 'data_tmp/docking/ligand1/ligand.pdb'
    from mbapy_lite.base import Configs
    b = Browser(options=[f"--user-agent={Configs.web.chrome_driver_path}"])
    run_dock_on_HPEPDOCK(receptor_path, ligand_path, browser=b)


'''
Date: 2024-08-18 13:26:20
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-23 19:55:05
Description: 
'''
import os
import uuid


def uuid4():
    """return a random 4-character string"""
    return uuid.uuid4().hex[:4]

def get_storage_path(sub_path:str):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'storage', sub_path)


'''
Date: 2024-05-06 17:40:28
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-23 20:21:30
Description: 
'''

from .__version__ import (__author__, __author_email__, __build__,
                          __description__, __license__, __title__, __url__,
                          __version__)


'''
Date: 2024-11-23 20:03:00
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-23 20:28:09
Description: 
'''
import argparse
import os
from typing import Dict, List

from mbapy_lite.file import get_paths_with_extension, opts_file

if __name__ == '__main__':
    from lazydock.gmx.thirdparty.sort_mol2_bonds import sort_bonds
    from lazydock.scripts._script_utils_ import clean_path, show_args
else:
    from ..gmx.thirdparty.sort_mol2_bonds import sort_bonds
    from ._script_utils_ import clean_path, show_args
    

def main(sys_args: List[str] = None):
    args_paser = argparse.ArgumentParser(description = 'Sort mol2 bonds by atom index.')
    args_paser.add_argument('-i', '--input', type = str, default='.',
                            help='input mol2 file or dir path. Default is %(default)s.')
    args_paser.add_argument('-s', '--suffix', type = str, default='_sorted',
                            help='suffix of sorted mol2 file name. Default is %(default)s.')
    args_paser.add_argument('-r', '--recursive', action='store_true', default=False,
                            help='FLAG, recursive search. Default is %(default)s.')
    args = args_paser.parse_args(sys_args)
    # process IO path
    args.input = clean_path(args.input)
    if os.path.isfile(args.input):
        paths = [args.input]
    elif os.path.isdir(args.input):
        paths = get_paths_with_extension(args.input, ['.mol2'], recursive=args.recursive)
    else:
        assert False, f'Input path {args.input} is not a file or dir.'
    # show args
    show_args(args, ['input', 'suffix', 'recursive'])
    # sort bonds
    for path in paths:
        print(f'Sorting bonds of {path} ...')
        sorted_mol2 = sort_bonds(opts_file(path))
        sorted_path = os.path.join(os.path.dirname(path), os.path.basename(path).split('.')[0] + args.suffix + '.mol2')
        opts_file(sorted_path, 'w', data=sorted_mol2)
        print(f'Sorted bonds saved to {sorted_path}.')


if __name__ == "__main__":
    main()


'''
Date: 2024-11-23 19:53:42
LastEditors: BHM-Bob 2262029386@qq.com
LastEditTime: 2024-11-23 19:53:56
Description:
'''
import argparse
import os
from pathlib import Path
from typing import Any, Dict, List, Union

from mbapy_lite.base import check_parameters_path, parameter_checker, put_err
from mbapy_lite.file import opts_file


def clean_path(path: str):
    return Path(path.replace('"', '').replace("'", '')).resolve()

def _print(content: str, f, verbose = True):
    if f is not None:
        f.write(content+'\n')
    if verbose:
        print(content)

def show_args(args, args_name: List[str], printf = print):
    printf('')
    for arg_name in args_name:
        printf(f'get arg: {arg_name}: {getattr(args, arg_name)}')
    printf('')

class Command:
    def __init__(self, args: argparse.Namespace, printf = print) -> None:
        self.args = args
        self.printf = printf
        self._pickle_except_list = []
        
    def process_args(self):
        pass
    
    def main_process(self):
        pass
        
    def excute(self):
        self.process_args()
        show_args(self.args, list(self.args.__dict__.keys()), self.printf)
        return self.main_process()
    
    def save_session(self, module_name: str, module_path: str = 'mbapy.scripts', path: str = os.curdir):
        if not Path(path).parent.exists():
            os.makedirs(Path(path).parent, exist_ok=True)
        session = {'__module_path__': module_path, '__module_name__': module_name, '__cmd_name__': self.__class__.__name__}
        for k,v in self.args.__dict__.items():
            if k not in self._pickle_except_list:
                session[k] = v
        opts_file(path, 'wb', way = 'pkl', data = session)
        
    @parameter_checker(path = check_parameters_path)
    def load_session(self, path: str):
        return opts_file(path, 'rb', way = 'pkl')
        
    def exec_from_session(self, session: Union[str, Dict[str, Any]]):
        if isinstance(session, str) and check_parameters_path(session):
            session = self.load_session(session)
        
    
def excute_command(args_paser: argparse.ArgumentParser, sys_args: List[str],
                   _str2func: Dict[str, callable]):
    args = args_paser.parse_args(sys_args)
    
    if args.sub_command in _str2func:
        try:
            if issubclass(_str2func[args.sub_command], Command):
                _str2func[args.sub_command](args).excute()
            else:
                _str2func[args.sub_command](args)
        except:
            if callable(_str2func[args.sub_command]):
                _str2func[args.sub_command](args)
    else:
        put_err(f'no such sub commmand: {args.sub_command}')
    
        
        
